﻿
local _G=_G
local strmatch,tonumber,format,debugprofilestop,table,tinsert,select,type,pairs,gsub,tostring,strlower = 
      strmatch,tonumber,format,debugprofilestop,table,tinsert,select,type,pairs,gsub,tostring,strlower
local max,next,issecurevariable = 
      max,next,issecurevariable
local EnumerateFrames, GetFrameCPUUsage,InCombatLockdown,GetFunctionCPUUsage,GetAddOnCPUUsage = 
	  EnumerateFrames, GetFrameCPUUsage,InCombatLockdown,GetFunctionCPUUsage,GetAddOnCPUUsage
	  
--GLOBALS: CPUThievesDB

local outputFrame = ChatFrame1
local addOnUpdateManually = false  -- we test later on if we need to add OnUpdate CPU usage ourselves (6.0.3: GetFrameCPUUsage doesn't include it!)

local printtmp = {"CPUThieves:"}
local function print(msgOrFrame,...)
	local frame
	local n=1
	if (type(msgOrFrame)=="table" and msgOrFrame.AddMessage) or msgOrFrame==nil then
		frame=msgOrFrame
	else
		n=n+1;printtmp[n]=msgOrFrame
	end
	if not frame then
		frame = _G.SELECTED_CHAT_FRAME or _G.DEFAULT_CHAT_FRAME
	end
	for i=1,select("#",...) do
		n=n+1;printtmp[n]=select(i,...)
	end
	local result = table.concat(printtmp," ",1,n);
	frame:AddMessage(result)
	
	local logs = CPUThievesDB.log
	logs.n = logs.n + 1
	if logs.n>1000 then
		logs.n=1
	end
	
	result = gsub(result, "|c[A-Fa-f0-9][A-Fa-f0-9][A-Fa-f0-9][A-Fa-f0-9][A-Fa-f0-9][A-Fa-f0-9][A-Fa-f0-9][A-Fa-f0-9]", "")
	result = gsub(result, "|r", "")
	logs[logs.n]=format("%s %s", _G.date("%H:%M:%S"), result)
end

local function printf(fmtOrFrame, ...)
	if (type(fmtOrFrame)=="table" and fmtOrFrame.AddMessage) or fmtOrFrame==nil then
		return print(fmtOrFrame,format(...))
	else
		return print(format(fmtOrFrame, ...))
	end
end


local isProfiling = GetScriptCPUUsage()>0	-- Is profiling enabled? (DON'T look at cvar, look at what's actually HAPPENING)


local pollAddOnCPU, pollFrameCPU, reportRecentCPU, initAddons, nowtolast
local outputStats

if isProfiling then
	local last = {}    -- ["AddonName"]=123 and [frame]=12
	local now = {}     -- ["AddonName"]=234 and [frame]=23
	local allAddons={} -- [1]="AddonName"

	local allFrames={}  -- [1]=frame
	
	initAddons = function()
		for i=1, GetNumAddOns() do
			local name = GetAddOnInfo(i)
			allAddons[i]=name
			now[name]=0
			last[name]=0
		end
	end
	

   
   -- GLOBALS: UpdateAddOnCPUUsage GetNumAddOns GetAddOnInfo  

   pollAddOnCPU = function(report)
		UpdateAddOnCPUUsage()
		for k,v in pairs(allAddons) do
			now[v]=GetAddOnCPUUsage(v)
			if not last[v] then
				last[v]=0
			elseif report and now[v]-last[v]>100 then
				if v~="!CPUThieves" then
					printf(outputFrame,
						" %5.0f |cff808080%4.1fs|r %s", now[v]-last[v], now[v]/1000, v)
				end
				last[v]=now[v]
			end
		end
	end
	 
	local function getFrameDisplayName(frame)
		local displayName = frame:GetName()
		if not displayName or displayName=="" then
			displayName = "UnnamedFrame" .. strlower(strmatch(tostring(frame), "table: *0*(.*)") or "?")
			_G[displayName]=frame -- so devs can go poking at it manually
			local parent = frame:GetParent()
			if parent then
				local pname = parent:GetName()
				if pname and pname~="" then
					displayName = format("%s Parent: %s", displayName, pname)
				end
			end
		end
		return displayName
	end
	
	local numFrames = 0
	local frame = nil
	local lastReportedFrameNow
	local numSame=0
	
	pollFrameCPU = function(report, pollAll)
		-- re-start polling sequence?
		if pollAll or not frame then
			frame = EnumerateFrames()
			numFrames=0
		end
		
		-- loop frames!
		while frame do
			numFrames=numFrames+1
			allFrames[numFrames]=frame
			
			local onUpdateCPU = 0
			if addOnUpdateManually then
				local func = frame:GetScript("OnUpdate")
				if func then 
					onUpdateCPU = GetFunctionCPUUsage(func, true) -- include called functions!
				end
			end
			
			now[frame]=GetFrameCPUUsage(frame,false) + onUpdateCPU
			
			if not last[frame] then
				last[frame] = 0
				-- Note that if this is a NEW frame attaching to a function ALREADY USED by 
				-- another frame for a while, this frame will randomly show as using lots 
				-- of CPU - once.
				-- Nothing to be done about it, really, unless we want to always skip measuring 
				-- new frames, which strikes me as an even worse idea.
			end
			
			if report and now[frame]-last[frame]>100 then
				-- Some LIMITED same-frame-handler resolution here. It only partially hides duplicates (due to wrongish sorting, and split scanning) but better than nothing.
				if lastReportedFrameNow == now[frame] then
					numSame = numSame + 1
				else
					if numSame>0 then
						printf(outputFrame,
							"    |cffc0c0c0 .. and %u more frames with same handler function|r", numSame)
						numSame=0
					end
					lastReportedFrameNow = now[frame]
					printf(outputFrame,
						" %5.0f |cff808080%4.1fs|r F:%s", now[frame]-last[frame], now[frame]/1000, getFrameDisplayName(frame))
				end
				-- IN CASE we end up re-scanning this frame before "now->last" copy (can happen!)
				-- we update the "last" value so we don't report the same timespan twice
				last[frame]=now[frame]	
			end
			
			-- Next!!
			frame = EnumerateFrames(frame)
			if not frame then
				-- Iterated all frames! break out of loop and "return true"
				break
			end
			
			-- Pause scanner via "return false" every 4000 frames (caller calls us again on next OnUpdate)
			if not pollAll and numFrames%4000 == 0 then
				return false
			end
		end
		if numSame>0 then	-- will only ever be >0 if reporting is enabled
			printf(outputFrame,
			"    |cffc0c0c0 .. and %u more frames with same handler function|r", numSame)
			numSame=0
		end
		return true
	end


   local function recentCpuComparator(a,b)
	  if a then
        a=now[a]-last[a]
	  else 
	    a=0
	  end
	  if b then
        b=now[b]-last[b]
	  else
	    b=0
	  end
      return a>b
   end
   
   local function totalCpuComparator(a,b)
	  if a then
        a=max(now[a] or 0, last[a] or 0)
	  else 
	    a=0
	  end
	  if b then
        b=max(now[b] or 0, last[b] or 0)
	  else
	    b=0
	  end
      return a>b
   end
   
	reportRecentCPU = function()
		table.sort(allAddons, recentCpuComparator)
		for i=1,5 do
			local name=allAddons[i]
			if now[name]-last[name] < 20 then break end -- dont output <20ms, they're not to blame
			printf(outputFrame,
				" %5.0f |cff808080%4.1fs|r %s", now[name]-last[name], now[name]/1000, name)
		end
		table.sort(allFrames, recentCpuComparator)
		local i=0
		for shown=1,5 do
			i=i+1
			local frame=allFrames[i]
			if now[frame]-last[frame] < 20 then break end -- dont output <20ms, they're not to blame
			-- find more frames with same handler func, which will be reported as ALSO taking the same time
			local numSame=0
			while now[allFrames[i+1]]==now[frame] do
				numSame=numSame+1
				i=i+1
			end
			
			printf(outputFrame,
				" %5.0f |cff808080%4.1fs|r F:%s", now[frame]-last[frame], now[frame]/1000, getFrameDisplayName(frame))
			if numSame>0 then
				printf(outputFrame,
				"    |cffc0c0c0 .. and %u more frames with same handler function|r", numSame)
			end
			
		end
		
   end
   
   
   local function totalCpuComparator(a,b)
	  if a then
        a=max(now[a] or 0, last[a] or 0)
	  else 
	    a=0
	  end
	  if b then
        b=max(now[b] or 0, last[b] or 0)
	  else
	    b=0
	  end
      return a>b
	end
   

	outputStats = function(numTop)
		numTop = numTop or 7
		printf("Tracking total %u addons, %u frames", #allAddons, #allFrames)
		table.sort(allAddons, totalCpuComparator)
		print("Addons by total CPU:")
		for i=1,numTop do
			local name=allAddons[i]
			printf(
				" |cffffff80%4.1fs|r %s", now[name]/1000, name)
		end
		table.sort(allFrames, totalCpuComparator)
		print("Frames by total CPU:")
		local i=0
		for shown=1,numTop do
			i=i+1
			local frame=allFrames[i]
			-- find more frames with same handler func, which will be reported as ALSO taking the same time
			local numSame=0
			while now[allFrames[i+1]]==now[frame] do
				numSame=numSame+1
				i=i+1
			end
			
			printf(
				" |cffffff80%4.1fs|r F:%s", now[frame]/1000, getFrameDisplayName(frame))
			if numSame>0 then
				printf(
				"    |cffc0c0c0 .. and %u more frames with same handler function|r", numSame)
			end
			
		end
	end   

	nowtolast = function()
		now,last = last,now	
	end
	
else
	pollAddOnCPU = function()
	end
	reportRecentCPU = function()
	end
	pollFrameCPU = function()
		return true	-- immediately done
	end
	initAddons = function()
	end
	nowtolast = function()
	end
	outputStats = function()
		print("CPU profiling is not enabled - no stats to show.")
	end
end



local frame = CreateFrame("Frame", "CPUThievesFrame")

local lastTime = debugprofilestop()
local lastScriptCPUUsage = GetScriptCPUUsage()
local pollingCycle=0
local myUsage = 0
local myUsageNum = 1
local myWorstUsage = 0

local reportOnNextFrame = false
local rebaseOnNextFrame = false

-- GLOBALS: GetScriptCPUUsage

local function onUpdate()
	local nowTime = debugprofilestop()
	local nowScriptCPUUsage = GetScriptCPUUsage()
	
	if rebaseOnNextFrame then
		rebaseOnNextFrame = false
		reportOnNextFrame = false
		-- Poll everything
		pollAddOnCPU(false)
		pollFrameCPU(false,true)
		-- Start polling normally again
		pollingCycle=0
		nowtolast()
	elseif nowTime-lastTime > (InCombatLockdown() and 150 or 300) then  -- UI errors out on 200+ ms in combat
		local hiccup = nowTime-lastTime
		local addons = nowScriptCPUUsage-lastScriptCPUUsage
		local r,g,b = 0xff,0xff,0xff
		if addons/hiccup > 0.2 then
		  b=32
		end
		if addons/hiccup > 0.7 then
		  g=32
		end
		printf(outputFrame,
			"Detected %.0fms FPS hiccup%s. (|cff%02x%02x%02x%.0fms = Lua|r)",
			hiccup,
			InCombatLockdown() and " DURING COMBAT" or "",
			r,g,b,
			addons
		)
		reportOnNextFrame = true	-- we can't report now, we'd hit the 200ms limit (if this is during combat)
		
	elseif reportOnNextFrame then
		reportOnNextFrame = false
		-- Poll everything
		pollAddOnCPU(false)
		pollFrameCPU(false,true)
		-- Report
		reportRecentCPU()
		-- Start polling normally again
		pollingCycle=0
		nowtolast()
	else
		if pollingCycle==0 then
			pollAddOnCPU(true)	-- true=WITH live reporting
			pollingCycle=1
		else
			local polledAll = pollFrameCPU(true)  -- true=WITH live reporting
			if polledAll then
				pollingCycle=0
				nowtolast()
			else
				pollingCycle=pollingCycle+1
			end
		end
		
		-- Do some stats on how much CPU CPUThieves itself eats (polling only)
		local thisUsage = debugprofilestop()-nowTime
		myUsage = myUsage + thisUsage
		myUsageNum = myUsageNum + 1
		if thisUsage > myWorstUsage then
			myWorstUsage = thisUsage
		end
	end
	
	-- Get ready for next pass
	lastTime=debugprofilestop()
	lastScriptCPUUsage = GetScriptCPUUsage()
	
end



local firstUpdate = true
frame:SetScript("OnUpdate", function()  -- only happens on 1 and 2nd OnUpdate, gets re-pointed
	if firstUpdate then
		firstUpdate = false
		-- 1st onupdate: initialization
		CPUThievesDB = CPUThievesDB or {}
		CPUThievesDB.log = CPUThievesDB.log or {}
		CPUThievesDB.log.n = CPUThievesDB.log.n or 0
		
		if type(_G[CPUThievesDB.chatFrame])=="table" and _G[CPUThievesDB.chatFrame].AddMessage then
			outputFrame = _G[CPUThievesDB.chatFrame]
		end

		initAddons()
		
		-- First, poll all current Addon & Frame CPU usages so we have a baseline
		pollAddOnCPU()
		pollFrameCPU(false, true)
		nowtolast()
		
		-- Use some CPU so we can do our test case on next call
		if isProfiling then
			local endTime=debugprofilestop() + 50
			while debugprofilestop() < endTime do
				for i=1,10000 do end	-- make sure THIS function takes most of the time, not API calls
			end
		end
	else
		-- 2nd onupdate: test if GetFrameCPUUsage() includes OnUpdate or not
		if isProfiling then
			-- print("FRAME CPU: ", GetFrameCPUUsage(frame))
			if GetFrameCPUUsage(frame)<30 then -- definitely didn't include the 50ms above
				addOnUpdateManually = true
			end
		end
			
		-- Move on to normal OnUpdates from here on out
		frame:SetScript("OnUpdate", onUpdate)
	end
end)
frame:Show()


local function quote(val)
	local typ = type(val)
	if typ=="number" then 
		return tostring(val)
	elseif typ=="string" then
		return format("%q", val)
	elseif typ=="table" or typ=="function" then
		return gsub(tostring(val), ": 0+", ": 0")	-- just chop out some leading 0s - there's MANY in 64bit!
	end
	return tostring(val)
end

local function identifyUnknown(unknown, skipname, tab, depth)
	tab = tab or _G
	depth = depth or 0
	-- depth of 6 takes about half a second on modern CPUs
	-- it incidentally also is deep enough to reach into e.g.
	--   Libstub.libs["LibName"].embeds["EmbedDestination"].memberVariable
	-- and
	--   DongleStub.versions["Astrolabe-1.0"].instance.processingFrame
	if depth>=6 then return end	
	
	for k,v in pairs(tab) do
		local res
		if k==unknown or v==unknown then
			if tab==_G and k==skipname then
				-- we already knew about this one, keep looking
			else
				return "["..quote(k).."] = "..quote(v)
			end
		elseif type(v)=="table" and v~=_G then
			res = identifyUnknown(unknown, skipname, v, depth+1)
			if res then
				return "["..quote(k).."]" .. res
			end
		elseif type(k)=="table" and k~=_G then
			res = identifyUnknown(unknown, skipname, k, depth+1)
			if res then
				return "[("..quote(k)..res..")] = "..quote(v)
			end
		end
	end
end



local allKnownScriptHandlers = {
  "OnAnimFinished",
  "OnAttributeChanged",
  "OnChar",
  "OnCharComposition",
  "OnClick",
  "OnColorSelect",
  "OnCursorChanged",
  "OnDisable",
  "OnDoubleClick",
  "OnDragStart",
  "OnDragStop",
  "OnEditFocusGained",
  "OnEditFocusLost",
  "OnEnable",
  "OnEnter",
  "OnEnterPressed",
  "OnEscapePressed",
  "OnEvent",
  "OnFinished",
  "OnHide",
  "OnHorizontalScroll",
  "OnHyperlinkClick",
  "OnHyperlinkEnter",
  "OnHyperlinkLeave",
  "OnInputLanguageChanged",
  "OnKeyDown",
  "OnKeyUp",
  "OnLeave",
  "OnLoad",
  "OnLoop",
  "OnMessageScrollChanged",
  "OnMinMaxChanged",
  "OnMouseDown",
  "OnMouseUp",
  "OnMouseWheel",
  "OnMovieFinished",
  "OnMovieHideSubtitle",
  "OnMovieShowSubtitle",
  "OnPause",
  "OnPlay",
  "OnReceiveDrag",
  "OnScrollRangeChanged",
  "OnShow",
  "OnSizeChanged",
  "OnSpacePressed",
  "OnStop",
  "OnTabPressed",
  "OnTextChanged",
  "OnTextSet",
  "OnTooltipAddMoney",
  "OnTooltipCleared",
  "OnTooltipSetAchievement",
  "OnTooltipSetDefaultAnchor",
  "OnTooltipSetEquipmentSet",
  "OnTooltipSetFrameStack",
  "OnTooltipSetItem",
  "OnTooltipSetQuest",
  "OnTooltipSetSpell",
  "OnTooltipSetUnit",
  "OnUpdate",
  "OnUpdateModel",
  "OnValueChanged",
  "OnVerticalScroll",
  "PostClick",
  "PreClick",
}



_G.SLASH_CPUTHIEVES1 = "/cputhieves"
_G.SLASH_CPUTHIEVES2 = "/cpu"
-- GLOBALS: SetCVar, ResetCPUUsage

SlashCmdList["CPUTHIEVES"] = function(arg)
	SELECTED_CHAT_FRAME:AddMessage("/cpu "..arg)
	if strmatch(arg,"^enable") then
		SetCVar("scriptProfile", 1)
		if isProfiling then
			print("Profiling is enabled.")
		else
			print("Reload UI to ENABLE PROFILING.")
		end
		
	elseif strmatch(arg,"^disable") then
		SetCVar("scriptProfile", 0)
		if not isProfiling then
			print("Profiling is disabled.")
		else
			print("Reload UI to DISABLE PROFILING.")
		end
		
		
		
	elseif strmatch(arg,"^output") then
	    local dest = strmatch(arg,"^output +([_A-Za-z0-9]+)")
		if type(_G[dest])=="table" and _G[dest].AddMessage then
			CPUThievesDB.chatFrame = dest
			outputFrame = _G[dest]
		else
			print("Usage: /cpu output ChatFrame#   (or anything with an .AddMessage)")
			-- GLOBALS: GetChatWindowInfo
			for i=1,9 do 
				local f=_G["ChatFrame"..i]
				local title=GetChatWindowInfo(i)
				if type(f)=="table" and f.AddMessage and title and title~="" then	
					print("  ChatFrame"..i..": "..GetChatWindowInfo(i))
				end
			end
		end
		
		
	elseif strmatch(arg,"^st[ats]*") then
		outputStats(tonumber(strmatch(arg, "^st[ats]* +([0-9]+)")))
		printf("CPUThieves uses about %.1fms per frame. Worst seen: %.1fms", (myUsage/myUsageNum), myWorstUsage)
		rebaseOnNextFrame = true
		
	elseif strmatch(arg, "^reset") then
		ResetCPUUsage();
		rebaseOnNextFrame = true
		
	elseif strmatch(arg,"^id[entify]* ") then
		local name = strmatch(arg,"^id[entify]* +(.*)") or ""
		local var = _G[name]
		if not var then
			printf("Can't find _G[%q]", name)
		else
			print("Trying to find variables pointing at "..quote(var))
			print("  "..(identifyUnknown(var, name) or "Can't."))
			if type(var)=="table" then
				if var.GetScript and var.HasScript then
					print("Trying to find variables pointing at frame's OnEvent, OnUpdate, etc")
					local found
					for _,scriptname in pairs(allKnownScriptHandlers) do
						local fn = var:HasScript(scriptname) and var:GetScript(scriptname)
						if fn then
							found = true
							print("  "..scriptname..": "..(identifyUnknown(fn, "") or "Can't."))
						end	
					end
					if not found then
						print("  No script handlers set.")
					end
				end
				print("Trying to issecurevariable() all members")
				local addons={}
				for k,v in pairs(var) do
					-- issecurevariable only reports back for string keys unfortunately
					local sec,who = issecurevariable(var,k)
					if not sec then
						addons[who] = (addons[who] or 0)+1
					end
				end
				if not next(addons) then
					print("  Can't find untainted.")
				else
					local t={}
					for name,count in pairs(addons) do
						t[#t+1] = format("%u x %q", count, name)
					end
					print("  "..table.concat(t, ", "))
				end
			end
		end
		rebaseOnNextFrame = true
	
	else
		print("Usage:")
		print("  enable - toggle profiling ON, needs reload")
		print("  disable - toggle profiling OFF, needs reload")
		print("  output - which chatframe to send live reporting to")
		print("  stats N - top N (default 7) CPU using addons&frames")
		print("  reset - reset all CPU usage counters")
		print("  identify VARNAME - try to identify who VARNAME (e.g. UnnamedFrame12345!) belongs to via scanning global variables and tables (up to 5 deep)")
		
		printf("CPU profiling is currently %s.", isProfiling and "ENABLED" or "DISABLED")
	end
end
